/*	CTE SIMULATOR, simulates the effects of CTE degradation of a CCD.
 *	VERSION: 0.0.2		(development release)
 *	LAST CHANGE: 7.Jan.2003
 *	AUTHOR: <bassler@phys.au.dk>
 *
 *  IMPORTANT NOTE: this code is still under heavy development just
 *	just the basic ideas are currently visible.
 *  Qualitative effects of CTE are o.k., but Quantitatively it is still not ok,
 *  and will not be so before release 0.1.0.
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <fitsio.h> /* fits reading routines */

#define DEFAULT_MODEL 0
#define GAIN 1		/* e / ADU */

/* other important model parameters */
/* Trap energies in MeV */
/* TODO: Put things in a .conf file (parse_config_file(filename)) */
#define ENERGY_PV 		0.44  		// currently only this trap is implemented
#define ENERGY_OV 		0.168
#define ENERGY_VV 		0.3

#define SIGMA_PV		3.5e-15		/* cm2 */
#define SIGMA_OV		000			// missing
#define SIGMA_VV		000			// missing

#define TEMPERATURE 	190			/* model temperature in Kelvin (not implemented yet)*/

#define BOLZMANNS_CONSTANT				1.3807e-23		/* J/K */
#define PLANCKS_CONSTANT				6.6262e-34		/* Js */
#define ELECTRON_MASS					9.102e-31		/* kg */
#define EFFECTIVE_ELECTRON_MASS_FACTOR	0.98			/* for Si */	
#define EV								1.6e-19			/* 1 eV = EV Joule */

#define ENTROPY_FACTOR_PV				1.43					
#define PI 3.1415927

#define OVERSCAN_Y 		25  	/* extra overscan rows option (useful for EPER)*/
#define FRAME_TRANSFER 	1 		/* frame transfer option. 0 = DISABLED, 1 = ENABLED */
#define T_SERIAL  		60 		/* serial CLK time in microseconds */
#define T_PARALLEL 		10000 	/* parallel CLK time in microseconds */
#define T_PARALLEL_FT 	1000	/* frame transfer CLK (if enabled) in microsecs */

/* MODEL 0:
	Dummy do-nothing mode for debugging purposes. 
	The CCD frame is only shifted and written as SIGNED_SHORT */

/* MODEL 1: 
   Simple CTE attenuation of signal.
   Needed Parameters: */
#define CTE_VALUE 0.992

/* MODEL 2:
	Advanced model by P. Bristow & A. Alexov (Oct 2002) 
	Needed Parameters: */

#define R_ZERO	0.28	/* fraction of traps exposed at zero signal */
#define N_T		1		/* density of bulk state traps (Depending on trap trype and NIEL) */
#define N_SAT	100000	/* Saturation signal level */

/* MODEL 3:
	New model 
	Needed parameters */
#define N_T		1	/* average trap density */

#define OUTPUT_FILE_NAME "out.fit"

void calc_probabilities(float time);
double calc_n_ex(float n_signal);
void **dyn2darray(unsigned row, unsigned col, unsigned el_size);
void rdfits(char *filename);
void wrfits(char *filename);
void printerror(int status);
void testfits0(void);			/* for debugging and testing */
void testfits1(void);			/* for debugging and testing */

float rd_pix(int x, int y);
int wr_pix(int x, int y, float f);
void init_rand(void);
double my_rand(double max);
int my_rand_test(float p);

int unsigned ccd_xsize;
int unsigned ccd_ysize;
int unsigned ccd_ysize_ft = 0;	/* frame transfer */
long int unsigned buffsize = 0;


float *frame_buffer;

float p_cn, p_en; 		/* capture and emission probabilites */

/* TODO: Code can be made faster by avoiding structures me thinketh */
typedef struct _pixel_type {	
		float  n_signal;
		double n_filled;
		int		traps;	/* the integer amount of traps in a pixel, used by model 3 */
} pixel_type;


int main(int argc, char *argv[])

{
	double n_pop = 0, n_unpop = 0;
	int i,j,k,l,x,y,t;
	int model = DEFAULT_MODEL;
	char filename[100];
	
	//float f;
	float **output_frame;

	pixel_type **calc_frame;
	double n_ex;
	float n_signal;
	float n_empty;
	float n_filled;
	float n_traps;
	
	long int total_ntraps;	/* model 3 */
	float n_signal_temp;
	float n_filled_temp;
	
	
	if (argc < 2 || argc > 3) {
		printf("Usage: %s [model] input_file.fit\n", argv[0]);
		exit(0);
	}
	
	if (argc == 2) {
		printf("   Using default model %i.", model);
		strcpy(filename,argv[1]);
	}
	
	if (argc == 3) {
		model = atoi(argv[1]);
		strcpy(filename,argv[2]);
		printf("   Using model %i\n", model);
	}
	
	/* TODO: a quick routine here for checking whether filename exists or not */
	/* TODO: Use getopt() instead */
	/* TODO: implement three phase subpixel structure */
	
	printf("   Reading data..\n");
	
	rdfits(filename);	 /* ccd_xsize and ccd_ysize is now known */
	//	testfits1();
	
	output_frame = (float **) dyn2darray(ccd_xsize, ccd_ysize + OVERSCAN_Y, sizeof(float));
	
	/* if it is a frame transfer CCD then the area is doubled */
	if (FRAME_TRANSFER)	
		ccd_ysize_ft = ccd_ysize;
	
	/* need a calc frame for holding the pixel parameters (need an extra empty y-row) 
	 */
	calc_frame = (pixel_type **) dyn2darray(ccd_xsize, ccd_ysize + ccd_ysize_ft + 1 + OVERSCAN_Y, sizeof(pixel_type));

	
	/* copy frame_buffer to calc_frame (ccd_ysize_ft is 0 if it is a normal CCD) */
	printf("   Initialization...\n"); /* init frame, transfer frame, overscan + 1 extra row */
	for (j = 0; j <= ccd_ysize + ccd_ysize_ft + OVERSCAN_Y; ++j) {
		for (i = 0; i < ccd_xsize; ++i) {
			calc_frame[i][j].n_signal = 0.0;
			calc_frame[i][j].n_filled = 0.0;
			calc_frame[i][j].traps = 0;
		}
	}
	
	/* randomize timer */
	init_rand();
	
	printf("   Copy frame_buffer to calc_frame...\n");
	for (j = 0; j < ccd_ysize; ++j) {
		for (i = 0; i < ccd_xsize; ++i) 
			calc_frame[i][j+ccd_ysize_ft].n_signal = rd_pix(i,j) * GAIN;		/* convert to #electrons */
	}
		
	printf("   Calculate things..\n");
	if (FRAME_TRANSFER) {
		calc_probabilities (T_PARALLEL_FT);   /* model 2 and 3 */
		printf("   Frame transfer mode enabled. FT Clock: %i\n", T_PARALLEL_FT);
	}
	else {
		printf("   Frame transfer mode disabled. Clock: %i\n", T_PARALLEL);
		calc_probabilities (T_PARALLEL);
	}
	
	printf("   P_pop: %f    P_unpop: %f.\n", p_cn, p_en);
	
	/* 	MODEL 3: generating some traps in calc_frame (NOT in the overscan area)*/
	total_ntraps = (int) (N_T * ccd_xsize * (ccd_ysize_ft + ccd_ysize) + 0.5);	/* calculate the total number of traps present */
	printf("   Now seeding %d traps...\n",(int) total_ntraps);
	/* Now set the number of traps in each pixel */
	for (i = 0; i < total_ntraps; ++i)	{
		x = (int) my_rand(ccd_xsize) + 0.5;
		y = (int) my_rand(ccd_ysize) + 0.5;
		++calc_frame[x][y].traps;
		// printf(" Seeding traps: %i %i %i\n", x, y, calc_frame[x][y].traps);
		// calc_frame[x][y].n_signal = 10;	/* make traps visible with a signal value */ /* DEBUGGER */
		}
	
	/* Increase ccd_ysize to include overscan area */
	ccd_ysize += OVERSCAN_Y;

	// seed a single trap for debugging:
	// calc_frame[25][24].n_signal = 107 ; // marker
	// calc_frame[25][24].traps = 10;
	
	/*  The following code may not be effective,
		but is written for transparency and clarity. (ho ho ho)*/
	/* TODO: Simulate serial shifts as well. */
	
	/* Simulate parallel shifts */
	printf("   Simulating charge shifts (this may take some time)...\n");
	for (k = 0; k < ccd_ysize + ccd_ysize_ft; ++k)
	{	
		
		if ((k == ccd_ysize) && FRAME_TRANSFER) {
			printf("   Frame transfer done! Calculating new probabilities for clock %i...\n",T_PARALLEL);
			calc_probabilities (T_PARALLEL);
			printf("   P_pop: %f    P_unpop: %f.\n", p_cn, p_en);
		}
		
		if (k >= ccd_ysize_ft) { /* frame transfer support */
			for (l = 0; l < ccd_xsize; ++l) 
				output_frame[l][k - ccd_ysize_ft] = calc_frame[l][0].n_signal / GAIN; /* copy (*,0) to output (*,y) and convert to ADU*/
		}
		/* Do ONE parallel shift */
		/* output amplifier located at pos (0,0) */	
		for (i = 0; i < ccd_ysize + ccd_ysize_ft; ++i) {
			for (j = 0; j < ccd_xsize; ++j)
			{
				/* shift next charge packet into this pixel and avoid timeconsuming access of array */
				n_signal = calc_frame[j][i+1].n_signal;
				n_filled = calc_frame[j][i].n_filled;
				n_traps = calc_frame[j][i].traps;
				/* get parameters ready */
									
				switch(model) {
					case 0:	/* do nothing dummy model */
						break;
					case 1:	/* simplest CTE model */
						n_signal *= CTE_VALUE;					
						break;
					case 2: /* P. Bristow & A. Alexov advanced CTE algorithm */
							/* TODO: algorithm still broken, this is only the outline */
						n_ex = calc_n_ex(n_signal);
						n_empty = n_ex - calc_frame[j][i].n_filled;
				
						if ( n_signal > n_empty)
							n_pop = n_empty * p_cn; /* p_cn is usually very close to 1 */
						else
							n_pop = n_signal * p_cn;
						n_unpop = -n_empty * p_en;		/* storage phase only */
						n_signal -= (n_pop - n_unpop);
				
						//if (k == 700)
						// printf("  %f",(float) (n_pop - n_unpop));
				
						calc_frame[j][i].n_filled = (n_pop - n_unpop);

						break;
					case 3:
						/* New model, check cap. and rel. charge, THEN update.*/
						/* will an electron be captured ? */
						n_filled_temp = n_filled;
						n_signal_temp = n_signal;
					
						if ((n_traps != 0) && (n_signal > 0)) {
							if (n_filled > n_traps) {
								printf("Oops-.. %i   ", (int) n_filled);
								printf("nf_t ns_t nt: %i %i %i\n", (int) n_filled_temp, (int) n_signal_temp, (int) n_traps);
								n_filled_temp = n_traps;
							}
							for( t = n_filled; (t < n_traps) && (n_signal_temp > 0); ++t ) {
								if (n_signal_temp < 1)
									printf(" hmmm: %f\n", n_signal_temp);
								if (my_rand_test(0.99)) {
									++n_filled_temp;  /* TODO: Check if signal is negative */
									--n_signal_temp;
									//if (k == 10) 
									// printf("captured %i   ", (int) n_filled);
								}
							}
						}
								
						/* will an electron be released ? */
						if (n_filled != 0) {
							if (n_filled < 1) {
								printf("Oops+..%i   ", (int) n_filled);
								n_filled = 0;
								}
							for (t = 0; t < n_filled; ++t)
								if (my_rand_test(0.9)) {		/* trap made very active */
									--n_filled_temp;
									++n_signal_temp;
									// if ( k == 10 )
									// printf("released. (fill signal t)  %i %i %i x:%i y:%i\n", (int) n_filled, (int) n_signal, t, j, i);
								}			
						}
						n_filled = n_filled_temp;
						n_signal = n_signal_temp;
						break;	
						
					default: /* do nothing */
						break;
						
						}		
					calc_frame[j][i].n_signal = n_signal;		
					calc_frame[j][i].n_filled = n_filled;		
					calc_frame[j][i].traps = n_traps;		
			}
		}
		// printf("  Reading out row number %d\n",k); /* hash */		
	}	


	/* frame_buffer size must be increased to hold overscan area */
	if (OVERSCAN_Y) {
		free(frame_buffer);
		frame_buffer = (float *) calloc (ccd_xsize * ccd_ysize, sizeof(float));	
	}
	
	/* copy output_frame to buffer */
	printf("   Copy output to frame_buffer...\n");
	for (i = 0; i < ccd_ysize; ++i) {
		for (j = 0; j < ccd_xsize; ++j)
			wr_pix(j,i,output_frame[j][i]);
	}
	
	free(calc_frame);
	free(output_frame);
	printf("   Writing data..\n");
	wrfits(OUTPUT_FILE_NAME);
	free(frame_buffer);	
	printf("   Done.\n");
	return (0);
}

void calc_probabilities(float time) {
	/* currently only using P-V valutes at -83 degC */
	/* TODO: include temperature dependancy. */
	/* TODO: include other traps. */
	// float capture_time; (depending on signal level, shit happens) 
	
	float release_time;
	float v_th;				/* thermal velocity */
	float n_c;				/* effective density of states */
	
	float m_e = ELECTRON_MASS * EFFECTIVE_ELECTRON_MASS_FACTOR;		/* effective electron mass (kg) */
	float kT = BOLZMANNS_CONSTANT * TEMPERATURE;	/* in J */
	
	
	/* Here the magic happens */
	v_th = sqrt( (3 * kT) / ( m_e ) );	/* (m/s) */
	v_th *= 100;	 					/* (cm/s) */
	n_c = 2 * pow( 2 * PI * m_e * kT / (PLANCKS_CONSTANT * PLANCKS_CONSTANT), 1.5 );
	release_time = exp(ENERGY_PV * EV / kT) / (SIGMA_PV * v_th * n_c);
	release_time *= 1e6;	/* convert to microseconds */
	
	printf ("  kT: %e\n", kT);
	printf ("   Thermal electron velocity (cm/s): %e -- N_c %e -- Rel.Time (s): %e\n", v_th, n_c, release_time);
	
	p_cn = 1 - exp(-time/0.047); /* in extreme cases p_cn is depending on the signal level */
	p_en = 1 - exp(-time/1.3e6);

}


double calc_n_ex(float n_signal)
{
	/* TODO: Monte-Carlo this. */
	/* TODO: Include clustering */
	/* TODO: If N_SIG > N_SAT then use alternative equation */
	
	double n_ex;
	n_ex = (double) (1-R_ZERO) * N_T * sqrt(n_signal/N_SAT) + R_ZERO * N_T;
	return (n_ex);
}

void **dyn2darray(unsigned row, unsigned col, unsigned el_size) {
	/*	Creates, allocates memory and returns a 2-dim array (row*col)
		where the elements have the size "el_size"
		The pointer must be freed after use. 
		(Code snippet from Bjarnes Thomsens IMSYS code.) */
	
	unsigned c;
	char *p = (char *)0;
	void **arr = (void **) calloc(
		row * sizeof(void *) + /* for pointers */
		row * col * el_size +  /* for elements */
		2 * sizeof(unsigned),  /* for dimensions */
		sizeof(char));

	if (arr == (void **)0)
		return arr;

	/* store row and col */
	p = (char *) arr + row * sizeof(void *);
	*(unsigned *) p = row;
	p += sizeof(unsigned);	/* p is a pointer to char */
	*(unsigned *) p = col;
	p += sizeof(unsigned);

	/* link rows to pointers */
	for (c = 0; c < row; ++c)
		arr[c] = p + (c * col * el_size);

	return arr;
}

void rdfits(char *filename)
    /************************************************************************/
    /* Read a FITS image into frame_buffer									*/
    /************************************************************************/
{
	/* float frame_buffer[ccd_xsize*ccd_ysize] is external */
	
    fitsfile *fptr;       /* pointer to the FITS file, defined in fitsio.h */
    int status,  nfound, anynull;
    long naxes[2], fpixel, nbuffer, npixels;

    float datamin, datamax, nullval;

    status = 0;

    if ( fits_open_file(&fptr, filename, READONLY, &status) )
         printerror(status);

    /* read the NAXIS1 and NAXIS2 keyword to get image size */
    if ( fits_read_keys_lng(fptr, "NAXIS", 1, 2, naxes, &nfound, &status) )
         printerror(status);

	ccd_xsize = naxes[0];	/* used externally -NB*/
	ccd_ysize = naxes[1];	
	
	printf("IMGSIZE: %d %d", ccd_xsize, ccd_ysize);
	
	npixels  = naxes[0] * naxes[1];         /* number of pixels in the image */
    fpixel   = 1;
    nullval  = 0;                /* don't check for null values in the image */
    datamin  = 1.0E30;
    datamax  = -1.0E30;

	buffsize = npixels;
	frame_buffer = (float *) calloc (buffsize, sizeof(float)); /* Will be freed later in main() */
	
    while (npixels > 0)
    {
      nbuffer = npixels;
      if (npixels > buffsize)
        nbuffer = buffsize;     /* read as many pixels as will fit in buffer */

      /* Note that even though the FITS images contains unsigned integer */
      /* pixel values (or more accurately, signed integer pixels with    */
      /* a bias of 32768),  this routine is reading the values into a    */
      /* float array.   Cfitsio automatically performs the datatype      */
      /* conversion in cases like this.                                  */

      if ( fits_read_img(fptr, TFLOAT, fpixel, nbuffer, &nullval,
                  frame_buffer, &anynull, &status) )
           printerror(status);

		  /* now "buffer" holds the image in float */
	  
      npixels -= nbuffer;    /* increment remaining number of pixels */
      fpixel  += nbuffer;    /* next pixel to be read in image */
    }

    printf("\nMin and max image pixels =  %.0f, %.0f\n", (float) datamin, (float) datamax);

    if ( fits_close_file(fptr, &status) )
         printerror(status);
    return;
}

void wrfits(char *filename)

    /******************************************************/
    /* Create a FITS primary array containing a 2-D image */
    /******************************************************/
{
    fitsfile *fptr;       /* pointer to the FITS file, defined in fitsio.h */
    int status, ii, jj;
    long  fpixel, nelements;/*, exposure;*/
    unsigned short *array[ccd_ysize]; 

    /* initialize FITS image parameters */
    int bitpix   =  SHORT_IMG; /* 16-bit signed short pixel values	*/
    long naxis    =   2;  /* 2-dimensional image					*/    
    long naxes[2] = { ccd_xsize, ccd_ysize };   /* image size */

    /* allocate memory for the whole image */ 
    array[0] = (unsigned short *)malloc( naxes[0] * naxes[1]
                                        * sizeof( unsigned short ) );

    /* initialize pointers to the start of each row of the image */
    for( ii=1; ii<naxes[1]; ii++ )
      array[ii] = array[ii-1] + naxes[0];

    remove(filename);               /* Delete old file if it already exists */

    status = 0;         /* initialize status before calling fitsio routines */

    if (fits_create_file(&fptr, filename, &status)) /* create new FITS file */
         printerror( status );           /* call printerror if error occurs */
    
    /* write the required keywords for the primary array image.     */
    /* Since bitpix = USHORT_IMG, this will cause cfitsio to create */
    /* a FITS image with BITPIX = 16 (signed short integers) with   */
    /* BSCALE = 1.0 and BZERO = 32768.  This is the convention that */
    /* FITS uses to store unsigned integers.  Note that the BSCALE  */
    /* and BZERO keywords will be automatically written by cfitsio  */
    /* in this case.                                                */

    if ( fits_create_img(fptr,  bitpix, naxis, naxes, &status) )
         printerror(status);          

	/* copy the frame_buffer into the array which will be written */
    for (jj = 0; jj < naxes[1]; jj++)
    {   for (ii = 0; ii < naxes[0]; ii++)
        {
            array[jj][ii] = frame_buffer[jj*ccd_xsize + ii];
        }
    }

	fpixel = 1;                               /* first pixel to write      */
    nelements = naxes[0] * naxes[1];          /* number of pixels to write */

    /* write the array of signed integers to the FITS file */ 						/* unsigned */
    if ( fits_write_img(fptr, TSHORT, fpixel, nelements, array[0], &status) )
        printerror(status);
      
    free( array[0]); /* free previously allocated memory */

    /* write another optional keyword to the header */
    /* Note that the ADDRESS of the value is passed in the routine */
	/* exposure = 1500.;
	if ( fits_update_key(fptr, TLONG, "EXPOSURE", &exposure,
		"Total Exposure Time", &status) )
		printerror( status );           */

	if ( fits_close_file(fptr, &status) )                /* close the file */
		printerror( status );           
}



void printerror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/
    if (status)
    {
       fits_report_error(stderr, status); /* print error report */

       exit( status );    /* terminate the program, returning error status */
    }
    return;
}


float rd_pix(int x, int y)
{
	float f;
	/* Read a pixel from the frame_buffer */
	if (x < ccd_xsize && y < ccd_ysize)
		f = frame_buffer[ccd_xsize * y + x];
	else
		f = 0;	/* return zero in case of out of area */
	return (f);
	
	}


int wr_pix(int x, int y, float f) 
{
	/* Write a pixel to the frame_buffer */
	frame_buffer[ccd_xsize * y + x] = f;
	return(0);
}


void init_rand(void) {
	/* randomize timer */
	time_t tp;
	
	tp = time(NULL);

	if (tp == -1)
		printf("WARNING: could not randomize timer.\n");
	else {
		printf("   Using seed: %i\n", (unsigned int) tp);
		srand((unsigned int) tp);
	}
}


double my_rand(double max) {
	/* This function returns a random value between 0 and max */
	
	double a;
	
	a = (double) (rand() * max) / RAND_MAX;
	return a;
}	


int my_rand_test(float p)
{
	int a;
	/* returns 1 (true) with the probability of p */
	if (p > my_rand(1))
		a = 1;
	else 
		a = 0;
	return(a);
}
		
void testfits0(void) {
	/* generates an empty frame */
	int ii, jj;
	
	ccd_xsize = 50;
	ccd_ysize = 50;
	
	frame_buffer = (float *) calloc (ccd_xsize * ccd_ysize, sizeof(float));
	
    for (jj = 0; jj < ccd_ysize; ++jj)
    {   for (ii = 0; ii < ccd_xsize; ++ii)
			frame_buffer[jj*ccd_xsize + ii] = 0;
    }	
	return;
	
}
	

void testfits1(void) {
	/* generates a frame with only one high pixel (1000) */
	int ii, jj;
	
	ccd_xsize = 50;
	ccd_ysize = 50;
		
	frame_buffer = (float *) calloc (ccd_xsize * ccd_ysize, sizeof(float));
	
    for (jj = 0; jj < ccd_ysize; ++jj)
    {   for (ii = 0; ii < ccd_xsize; ++ii) 
			frame_buffer[jj*ccd_xsize + ii] = 0;
    }	
	
	frame_buffer[25 * 50 + 25 ] = 100;
	
	return;
	
}
